Require TransitionSystems.

Definition message := nat.

Inductive action : Set :=
   In : message -> action
 | Rd : message -> action
 | Out : message -> action
 | BangIn : message -> action.

Inductive program : Set :=
   Dead : program
 | Action : action -> program -> program
 | IfThenElseAction : action -> program -> program -> program
 | Parallel : program -> program -> program.

Inductive agent : Set :=
   Message : message -> agent
 | Program : program -> agent
 | AParallel : agent -> agent -> agent.

Inductive label : Set :=
   Tau : label
 | LIn : message -> label
 | LRead : message -> label
 | LOut : message -> label
 | LNot : message -> label.

(* This predicate states that no message m is present in the Tuple Space *)
Inductive NoMessage [m:message] : agent -> Prop :=
   NMMessage :
    (m':message)
     ~m=m' -> (NoMessage m (Message m'))
 | NMProgram : (P:program)(NoMessage m (Program P))
 | NMParallel :
    (P,Q:agent)
     (NoMessage m P) -> (NoMessage m Q) -> (NoMessage m (AParallel P Q)).

(* A step in the ORDERED semantics. *)
Inductive OStep : agent -> label -> agent -> Prop :=
   SMessage : (m:message)(OStep (Message m) (LOut m) (Program Dead))
   (* This is the transition that makes the semantics ORDERED *)
 | SOut :
    (m:message; P:program)
     (OStep
       (Program (Action (Out m) P))
       Tau
       (AParallel (Message m) (Program P)))
 | SIn :
    (m:message; P:program)
     (OStep (Program (Action (In m) P)) (LIn m) (Program P))
 | SRead :
    (m:message; P:program)
     (OStep (Program (Action (Rd m) P)) (LRead m) (Program P))
 | SBangIn :
    (m:message; P:program)
     (OStep
       (Program (Action (BangIn m) P))
       (LIn m)
       (AParallel (Program P) (Program (Action (BangIn m) P))))
 | SInpT :
    (m:message; P,Q:program)
     (OStep (Program (IfThenElseAction (In m) P Q)) (LIn m) (Program P))
 | SRdpT :
    (m:message; P,Q:program)
     (OStep (Program (IfThenElseAction (Rd m) P Q)) (LRead m) (Program P))
 | SInpF :
    (m:message; P,Q:program)
     (OStep (Program (IfThenElseAction (In m) P Q)) (LNot m) (Program Q))
 | SRdpF :
    (m:message; P,Q:program)
     (OStep (Program (IfThenElseAction (Rd m) P Q)) (LNot m) (Program Q))
 | SAParallelPL :
    (P,P',Q:agent; l:label)
     ~(EX m | l = (LNot m)) ->
      (OStep P l P') -> (OStep (AParallel P Q) l (AParallel P' Q))
 | SAParallelPR :
    (P,P',Q:agent; l:label)
     ~(EX m | l = (LNot m)) ->
      (OStep P l P') -> (OStep (AParallel Q P) l (AParallel Q P'))
 | SAParallelNL :
    (P,P',Q:agent; m:message)
     (OStep P (LNot m) P') ->
      (NoMessage m Q) ->
       (OStep (AParallel P Q) (LNot m) (AParallel P' Q))
 | SAParallelNR :
    (P,P',Q:agent; m:message)
     (OStep P (LNot m) P') ->
      (NoMessage m Q) ->
       (OStep (AParallel Q P) (LNot m) (AParallel Q P'))
 | SAParallelTauIn1 :
    (P,P',Q,Q':agent; m:message)
     (OStep P (LIn m) P') ->
      (OStep Q (LOut m) Q') ->
       (OStep (AParallel P Q) Tau (AParallel P' Q'))
 | SAParallelTauIn2 :
    (P,P',Q,Q':agent; m:message)
     (OStep P (LOut m) P') ->
      (OStep Q (LIn m) Q') ->
       (OStep (AParallel P Q) Tau (AParallel P' Q'))
 | SAParallelTauRead1 :
    (P,P',Q,Q':agent; m:message)
     (OStep P (LRead m) P') ->
      (OStep Q (LOut m) Q') ->
       (OStep (AParallel P Q) Tau (AParallel P' Q'))
 | SAParallelTauRead2 :
    (P,P',Q,Q':agent; m:message)
     (OStep P (LOut m) P') ->
      (OStep Q (LRead m) Q') ->
       (OStep (AParallel P Q) Tau (AParallel P' Q'))
 | SParallel :
    (P,Q: program)
      (OStep (Program (Parallel P Q)) Tau (AParallel (Program P) (Program Q))).

(* Step is neither deterministic nor total *)
Definition step :=
 [P,P':agent]
  (OStep P Tau P') \/
   (EX m | (OStep P (LNot m) P')).

Definition step_plus :=
 (TransitionSystems.step_plus ? step).

Definition step_star :=
 (TransitionSystems.step_star ? step).

Definition stuck :=
 (TransitionSystems.stuck ? step).

(* reduces is neither deterministic nor total *)
Definition reduces :=
 (TransitionSystems.reduces ? step).
